@@ -11,7 +11,7 @@ def avg_exp(x):
         value
     """
     with np.errstate(invalid="ignore"):
-        rv = np.where(np.abs(x) < 1e-5,
-                      1 + (x/2)*(1 + (x/3)*(1 + (x/4))),
+        rv = np.where(np.abs(x) < 1e-6,
+                      1. + (x/2.)*(1. + (x/3.)),  # + o(x^3)
                       np.expm1(x)/x)
     return rv
